/* $Id: watch.c,v 1.15 2008/06/28 06:00:29 harding Exp $ */

/*
 * Copyright (c) Carson Harding, 2005. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/event.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ctype.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <libgen.h>
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#define SYSLOG_NAMES	/* get syslog facility and priority names */
#include <syslog.h>


/* 
 * Pulled from sys/proc.h behind if _KERNEL; nice if this was 
 * accessible from include or sysctl().
 */
#define	PID_MAX		32766

/*
 * Lookups for names of events and descriptions
*/
struct	event_values {
	char	*name;
	int	value;
	char	*descr;
};

struct 	event_values	vnode_ev[] = {
	{ "delete" ,  NOTE_DELETE ,   "link deleted" },
	{ "write" ,   NOTE_WRITE ,    "written to" },
	{ "extend" ,  NOTE_EXTEND ,   "extended" },
	{ "truncate", NOTE_TRUNCATE , "truncated" },
	{ "attrib" ,  NOTE_ATTRIB ,   "attribute change" },
	{ "link" ,    NOTE_LINK ,     "link count changed"},
	{ "rename" ,  NOTE_RENAME ,   "renamed" },
	{ "revoke" ,  NOTE_REVOKE ,   "revoked" }
};
int	vnode_en = (sizeof(vnode_ev)/sizeof(struct event_values));

struct	event_values	proc_ev[] = {
	{ "exit" ,    NOTE_EXIT ,     "process exited" },
	{ "fork" ,    NOTE_FORK ,     "process forked" },
	{ "exec" ,    NOTE_EXEC ,     "process exec'd another" }
};
int	proc_en	= (sizeof(proc_ev)/sizeof(struct event_values));

extern	char	*__progname;

int	f_quiet;	/* don't emit any messages about event */
int	f_time;		/* prefix event message with time */
int	f_continue;	/* don't exit after event */
int	f_modify;	/* watch content modification when EVFILT_VNODE */
int	f_pollopen;	/* poll on file creation if not present */
int	f_syslog;	/* use syslog, not stdout */
int	f_facility = LOG_USER;
int	f_priority = LOG_INFO;
int	f_daemon;	/* drop to background */

void	usage(void);
void	setup_syslog(char *spec, int *facility, int *priority);
int	event_flags(char *events, int filter);
int	poll_open(char *file);
void	check_fs(int fd);
void	watch(char *identstr, int ident, int filter, int fflags);
char	*decode_event(struct kevent *ke);
void	print_event(char *identstr, char *evstr);
int	parse_events(char *s, int eventsn, struct event_values *eventsv);
int	event_value(char *s, int en, struct event_values *ev);

__dead void
usage(void)
{
	int	n;

	fprintf(stderr, 
	    "usage: %s "
	    "{-p pid | -f file [-m]} "
	    "[-cnqt] [-S [facility.priority]] "
	    "[-e event[,event...]]\n",
	    __progname);
	fprintf(stderr, "\n"
	    "    -f   - watch a file or directory\n"
	    "    -p   - watch a process\n"
	    "    -c   - continuously watch (even across file deletion)\n"
	    "    -d   - drop to background\n"
	    "    -m   - add data changes to file events to watch for\n"
	    "    -n   - if file does not exist, wait for it to exist\n"
	    "    -q   - quiet (incompatible with -c)\n"
	    "    -t   - include time in event message\n"
	    "    -S   - use syslog, optionally specifying the facility \n"
	    "           and priority\n"
	    "    -e   - watch for specified events. Valid event names are\n"
	    "           for -f:\n");
	fprintf(stderr, "              ");
	for (n = 0; n < vnode_en-1; n++)
		fprintf(stderr, " %s,", vnode_ev[n].name);
	fprintf(stderr, " %s\n", vnode_ev[n].name);

	fprintf(stderr, "           for -p:\n");
	fprintf(stderr, "              ");
	for (n = 0; n < proc_en-1; n++)
		fprintf(stderr, " %s,", proc_ev[n].name);
	fprintf(stderr, " %s\n", proc_ev[n].name);

	exit(1);
}

int
main(int argc, char **argv)
{
	int	ch;
	int	ident = -1;
	int	filter = 0;
	int	fflags = 0;
	char	*events = NULL;
	char	*identstr = NULL;
	const	char *errstr = NULL;

	while ((ch = getopt(argc, argv, "cde:f:mnp:qtS::V")) != -1) {
		switch(ch) {
		case 'c':
			f_continue = 1;
			break;
		case 'd':
			f_daemon = 1;
			break;
		case 'e':
			events = optarg;
			break;
		case 'f':
			if (filter)
				usage();
			identstr = optarg;
			filter = EVFILT_VNODE;
			break;
		case 'm':
			f_modify = 1;
			break;
		case 'n':
			f_pollopen = 1;
			break;
		case 'p':
			if (filter)
				usage();
			identstr = optarg;
			ident = (int)strtonum(
				    identstr, 1, PID_MAX, &errstr);
			if (errstr)
				errx(1, "%s", errstr);
			filter = EVFILT_PROC;
			break;
		case 'q':
			f_quiet = 1;
			break;
		case 't':
			f_time = 1;
			break;
		case 'S':
			f_syslog = 1;
			setup_syslog(optarg, &f_facility, &f_priority);
			break;
		case 'V':
			fprintf(stdout, "%s %s\n", __progname, VER);
			exit(0);
			break;
		default:
			usage();
			break;
		}
	}

	if (!filter)
		usage();

	if (f_continue && f_quiet)
		errx(1, "-q and -c are incompatible flags");

	fflags = event_flags(events, filter);

	if (f_syslog)
		openlog(__progname, LOG_PID|LOG_NDELAY, f_facility);
	if (f_daemon)
		daemon(0, 0);

	if (filter == EVFILT_VNODE) {
		/* get file handle, check file is on local, BSD fs */
		do {
			ident = open(identstr, O_RDONLY, 0);
			if (ident == -1) {
				if (errno == ENOENT && f_pollopen) {
					ident = poll_open(identstr);
					/* 
					 * Existing when didn't exist before
					 * counts as an event. poll_open()
					 * never returns if can't open.
					 */
					if (!f_continue)
						break;
				} else
					err(1, "unable to open %s", identstr);
			}

			/* make sure on supported file system */
			check_fs(ident);

			watch(identstr, ident, filter, fflags);

			(void)close(ident);

			/*
			 * If continue flag is set, then this implies
			 * we are to poll to open after the initial run.
			 * We may require the file to be there at start,
			 * but we also want to continue to look for it if
			 * it now should go away.
			 */
			if (f_continue)
				f_pollopen = 1;

		} while (f_continue);
	} else 
		watch(identstr, ident, filter, fflags);

	exit(0);
}

void
setup_syslog(char *spec, int *facility, int *priority)
{
	char	*f, *p;
	int	fn = 0, pn = 0;
	CODE	*cdp;

	if (!spec)
		return;
	
	f = strdup(spec);
	if (!f)
		err(1, NULL);

	if ((p = strchr(f, '.')) == NULL)
		errx(1, "no \".\" in syslog configuration string");
	*p++ = '\0';

	for (cdp = facilitynames; cdp->c_name != NULL; cdp++) {
		if (strcasecmp(cdp->c_name, f) == 0) {
			fn = cdp->c_val;
			break;
		}
	}
	if (!cdp->c_name)
		errx(1, "%s: invalid facility name", f);

	for (cdp = prioritynames; cdp->c_name != NULL; cdp++) {
		if (strcasecmp(cdp->c_name, p) == 0) {
			pn = cdp->c_val;
			break;
		}
	}
	if (!cdp->c_name)
		errx(1, "%s: invalid priority name", p);

	*facility = fn;
	*priority = pn;
}

/* 
 * Set event filter flags, either as specified in the events string,
 * or to defaults if events string is NULL. For vnode, by default we 
 * look for inode changes and not data changes. We add data changes if
 * the command-line flag to do so was set, or, of course, if they
 * are in the the list of user-specified events. For proc, by
 * default we look for any changes, and don't set track.
 *
 * If the user is specifying the flags:
 *	- We always add NOTE_DELETE; else we could be left with
 *	  an open descriptor to something nothing is ever going
 *	  to happen to again;
 *	- Similarily, we always add NOTE_EXIT when watching a 
 *	  process. The process is gone: nothing to see.
 */
int
event_flags(char *events, int filter)
{
	int	fflags;

	if (filter == EVFILT_VNODE) {
		if (events)
			fflags = parse_events(events, vnode_en, vnode_ev);
		else {
			fflags = NOTE_ATTRIB | NOTE_LINK | NOTE_RENAME | 
			    NOTE_REVOKE;
			if (f_modify)
				fflags |= NOTE_WRITE | NOTE_EXTEND | 
				    NOTE_TRUNCATE;
		}
		fflags |= NOTE_DELETE;
	} else {
		if (events)
			fflags = parse_events(events, proc_en, proc_ev);
		else
			fflags = NOTE_FORK | NOTE_EXEC;
		fflags |= NOTE_EXIT;
	}

	return fflags;
}

/*
 * If we can't open the file, we find the parent directory and sit and
 * watch it for changes; when it changes, we try to open the file again.
 * We can then call this again if the file is deleted and we want to watch
 * for it to reappear.
 */
int
poll_open(char *file)
{
	int	fd, dirfd;
	char 	fpath[PATH_MAX], *dpath;
	int	kq, kval;
	struct	kevent ke;

	/* 
	 * Get real path and directory. Note that OpenBSD's implementation 
	 * does not return ENOENT if the final component does not exist.
	 */
	if (!realpath(file, fpath))
		err(1, "%s", fpath);

	dpath = dirname(fpath);
	dirfd = open(dpath, O_RDONLY, 0);
	if (dirfd == -1)
		err(1, "unable to open directory: %s", dpath);

	kq = kqueue();
	if (kq == -1)
		err(1, "error getting kqueue descriptor");

	 /* Look for write events, but also need to handle delete,revoke. */
	EV_SET(&ke, dirfd, 
	    EVFILT_VNODE, 
	    EV_ADD|EV_CLEAR,
	    NOTE_WRITE|NOTE_DELETE|NOTE_REVOKE,
	    0, NULL);

	if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0)
		err(1, "unable to register kevent");
	
	for (fd = -1; fd == -1; ) {
		fd = open(file, O_RDONLY, 0);
		if (fd == -1)  {
			if (errno != ENOENT)
				err(1, "error opening %s", file);

			if ((kval = kevent(kq, NULL, 0, &ke, 1, NULL)) < 0)
				errx(1, "kevent error");
			if ((ke.fflags & NOTE_DELETE) 
			    || (ke.fflags & NOTE_REVOKE))
				errx(1, "can no longer access path to %s", file);
		}
	}

	print_event(file, "now exists");

	(void)close(kq);
	(void)close(dirfd);

	return fd;
}

/*
 * Check the filesystem is a supported type; exit if not.
 */
void
check_fs(int fd)
{
	struct	statfs	fsb;

	if (fstatfs(fd, &fsb) == -1)
		err(1, "unable to stat file system");
	if (strcmp("ffs", fsb.f_fstypename) != 0 &&
	    strcmp("mfs", fsb.f_fstypename) != 0) {
		errx(1, "file must be on local BSD file system");
	}
}

/*
 * Watch the file or process, handle events.
 */
void
watch(char *identstr, int ident, int filter, int fflags)
{
	int	kq, kval, flags;
	struct	kevent	ke;
	struct	stat	sb;
	int	stop = 0;

	flags = EV_ADD|EV_CLEAR;
	if (!f_continue)
		flags |= EV_ONESHOT;

	kq = kqueue();
	if (kq == -1)
		err(1, NULL);

	EV_SET(&ke, ident, filter, flags, fflags, 0, NULL);

	if (kevent(kq, &ke, 1, NULL, 0, NULL) < 0)
		err(1, "unable to register kevent for %d", ident);

	while ( !stop ) {
		kval = kevent(kq, NULL, 0, &ke, 1, NULL);
		if (kval < 0)
			err(1, "kevent error");

		if (filter == EVFILT_VNODE) { 
			if ((ke.fflags & NOTE_DELETE)) {
				if (fstat(ke.ident, &sb) == -1)
					err(1, "fstat failed");
				/* if last link, then bail */
				if (sb.st_nlink < 1)
					stop = 1;
			}
			if ((ke.fflags & NOTE_REVOKE))
				stop = 1;
		} 
		else if ((ke.fflags & NOTE_PCTRLMASK) == NOTE_EXIT)
			stop = 1;

		if (!f_quiet)
			print_event(identstr, decode_event(&ke));
		if (!f_continue)
			stop = 1;

	}

	(void)close(kq);
}

/*
 * Decode an event into descriptive message.
 */
char *
decode_event(struct kevent *ke)
{
	int	n, evn;
	struct	event_values *ev;

	switch(ke->filter) { 
	case EVFILT_VNODE:
		ev = vnode_ev;
		evn = vnode_en;
		break;
	case EVFILT_PROC:
		ev = proc_ev;
		evn = proc_en;
		break;
	default:
		return "invalid filter";
	}

	for (n = 0; n < evn; n++) {
		if ((ke->fflags & ev[n].value))
			return ev[n].descr;
	}

	return "unknown event";
}

/*
 * Print formatted message.
 */
void
print_event(char *identstr, char *evstr)
{
	time_t	now;
	struct	tm *tm;	
	char	timebuf[32];

	if (f_syslog)
		syslog(f_priority, "%s: %s", identstr, evstr);
	else if (!f_daemon) {
		fprintf(stdout, "%s: ", __progname);
		if (f_time) {
			(void)time(&now);
			tm = localtime(&now);
			strftime(timebuf, sizeof(timebuf), "%c", tm);
			fprintf(stdout, "%s ", timebuf);
		}
		fprintf(stdout, "%s: %s.\n", identstr, evstr);
	}
}

/*
 * Events list comes in as a comma-separated list of event names.
 * We chop the event string and lookup filter values in tables, and
 * Return the OR'd set of flags.
 */
int
parse_events(char *events, int en, struct event_values *ev)
{
	int	e, eflags = 0;
	char	*s, *ss;
	
	s = strdup(events);
	if (!s)
		return -1;

	while ((ss = strsep(&s, ", ")) != NULL) {
		if (*ss == '\0')
			continue;
		if ((e = event_value(ss, en, ev)) == -1)
			errx(1, "%s: not a valid event for filter type", ss);
		eflags |= e;
	}

	free(s);

	return eflags;
}

/*
 * Return the value referred to by string s, or -1 if it
 * does not exist in table.
 */
int
event_value(char *s, int en, struct event_values *ev)
{
	int	n;

	for (n = 0; n < en; n++) {
		if (strcasecmp(ev[n].name, s) == 0) {
			return ev[n].value;
		}
	}

	return -1;
}
